'use strict';

/* jslint indent: 4 */

function translateKey(key) {
    return key.replace(/_/g, '-');
}

// translates the boolean-typed value of a JSON environment option to process arguments
function translateJSONEnvBooleanOptionToOption(logger, options, name, key, argv) {
    if (!options.hasOwnProperty(key)) {
        return;
    }

    var option = options[key];
    if (option === true) {
        argv.push('--' + translateKey(key));
    } else if (option === false) {
        argv.push('--no-' + translateKey(key));
    } else {
        logger.error('Option "' + key + '" in ' + name + ' needs to have a boolean value');
        process.exit(1);
    }
}

// translates the integer-typed value of a JSON environment option to process arguments
function translateJSONEnvIntegerOptionToOption(logger, options, name, key, argv) {
    if (!options.hasOwnProperty(key)) {
        return;
    }

    var option = options[key];
    if (Number.isInteger(option)) {
        argv.push('--' + translateKey(key));
        argv.push(option);
    } else {
        logger.error('Option "' + key + '" in ' + name + ' needs to have an integer value');
        process.exit(1);
    }
}

// translates the string-typed value of a JSON environment option to process arguments
function translateJSONEnvStringOptionToOption(logger, options, name, key, argv) {
    if (!options.hasOwnProperty(key)) {
        return;
    }

    var option = options[key];
    if (typeof option === "string") {
        argv.push('--' + translateKey(key));
        argv.push(option);
    } else {
        logger.error('Option "' + key + '" in ' + name + ' needs to have a string value');
        process.exit(1);
    }
}

// translates the string-array-typed value of a JSON environment option to process arguments
function translateJSONEnvStringArrayOptionToOption(logger, options, name, key, argv) {
    if (!options.hasOwnProperty(key)) {
        return;
    }

    var option = options[key];
    if (Array.isArray(option)) {
        for (var i = 0; i < option.length; ++i) {
            var path = option[i];
            if (typeof path !== "string") {
                logger.error('Option "' + key + '" in ' + name + ' needs to have a string value at array element ' + i);
                process.exit(1);
            }
        }
        argv.push('--' + translateKey(key));
        argv.push(option);
    } else {
        logger.error('Option "' + key + '" in ' + name + ' needs to have an array value');
        process.exit(1);
    }
}

// translates the given string list-option via the given callback function
function translateStringListOption(i, argv, cb) {
    ++i;
    var nextArgIsNotAnOption = false;
    while (i < argv.length) {
        var arg = argv[i];
        if (typeof arg === 'string' && arg !== '') {
            if (nextArgIsNotAnOption) {
                // this is definitely a value
                cb(arg);
                nextArgIsNotAnOption = false;
            } else {
                if (arg === '--') {
                    // it's a marker to allow values which start with --
                    nextArgIsNotAnOption = true;
                } else if (arg.length > 2 && arg[0] === '-' && arg[1] === '-') {
                    // it's an option, push it back and continue in caller('s loop)
                    --i;
                    break;
                } else {
                    // it's a value
                    cb(arg);
                }
            }
        } else if (Array.isArray(arg)) {
            // it's a list of values
            for (var j = 0; j < arg.length; ++j) {
                var argj = arg[j];
                if (typeof argj === 'string' && argj !== '') {
                    cb(argj);
                }
            }
        }
        ++i;
    }
    return i;
}

// translates the given string list-option option to an array
function translateStringListOptionToArray(i, argv, array) {
    while (array.length > 0) {
        array.pop();
    }
    return translateStringListOption(i, argv, function(value) {
        array.push(value);
    });
}

// translates the given path-filter-like option to a path filter object, e.g. --include-filter path1 path2 -- --path3
function translatePathListOptionToPathFilter(i, argv, pathFilter, cb) {
    pathFilter.invalidate();
    return translateStringListOption(i, argv, function(path) {
        if (cb) {
            cb(path);
        }
        pathFilter.addPath(path);
    });
}

module.exports = {
    translateJSONEnvBooleanOptionToOption: translateJSONEnvBooleanOptionToOption,
    translateJSONEnvIntegerOptionToOption: translateJSONEnvIntegerOptionToOption,
    translateJSONEnvStringOptionToOption: translateJSONEnvStringOptionToOption,
    translateJSONEnvStringArrayOptionToOption: translateJSONEnvStringArrayOptionToOption,
    translateStringListOptionToArray: translateStringListOptionToArray,
    translatePathListOptionToPathFilter: translatePathListOptionToPathFilter
};
