'use strict';

const coFs = require('fs');
const coCrypto = require('crypto');
const coUtils = require('../utils.js');
const Result = require('./result.js').Result;
const coDwfAbstract = require('./abstract.js');

const coAsyncLib = require('async');
const coPathLib = require('../paths.js');
const coPathsLib = require('../paths.js');
const coLoggerLib = require('../logger.js');
const coValidateLib = require('./validate.js');
const coTxtLib = require('./object/type/txt.js');
const coAbstractLib = require('./object/type/abstract.js');
const coTaskchainLib = require('./object/type/taskChain.js');
const coNamespaceLib = require('./object/type/nameSpace.js');
const coDLMProfileLib = require('./object/type/dlmProfile.js');

const cfGetHash = function (ioTaskChain4Batch) {
  let lsText = JSON.stringify(ioTaskChain4Batch);
  let lsHash = coCrypto.createHash('sha256').update(lsText, coUtils.encoding).digest('hex');
  return lsHash;
};
const cfGetFile4BatchPost = function (isFile, ioDir2Namespace) {
  let loData_TaskChain = coUtils.readJSONFile(isFile);
  let laParsedNamelessFile = coPathsLib.parseNamelessFile(isFile);
  let lsNamespace = ioDir2Namespace[laParsedNamelessFile.dir]; // Check
  if (typeof lsNamespace === 'undefined' || lsNamespace === null) {
    throw new Error('Failed, because the \'' + '\' namespace for ' + isFile + ' file is not supported');
  }
  let lsName = laParsedNamelessFile.name;
  if (typeof lsName === 'undefined' || lsName === null) { // Check - .dwftaskchain
    throw new Error('Failed, because the \'' + '\' id for ' + isFile + ' file is not supported');
  }
  loData_TaskChain[coAbstractLib.coColumn[laParsedNamelessFile.ext]] = lsName;
  loData_TaskChain[coAbstractLib.csKey_AOI_csNamespace] = lsNamespace;
  return loData_TaskChain;
};

/*
* Receiving credentials, list of all files in the design time dir.
* iaoFilesRepo(input array which has objects in it) is an object which in fact contains response of Get call on url/repo/objects
* iaoFilesRepo [{ID, NAMESPACE, TYPE_ID(.dwftaskchain) , DESIGN_TIME_FILE_PATH, HASH_UTF8_SHA256_HEX}] till today
*/
function Container(ioCredential, ioOption, iaoFile, iaoFilesRepo) {
  this.loCredential = coValidateLib(ioCredential, 'object'); // Check - ioCredential schema validation
  this.loOption = coValidateLib(ioOption, 'object');
  this.laoFileDesign = coValidateLib(iaoFile, 'object');
  this.laoFilesRepo = coValidateLib(iaoFilesRepo[coAbstractLib.TYPES.OBJECTS], 'object');
  this.loProviders = { 'POST': [], 'DELETE': [] };
  this.loObjectType = { 'POST': [], 'DELETE': [] };

  let lnVersion = ioCredential[coDwfAbstract.cs_Credential_Key_Version];
  let lsVersionErrorMessage =
    'Failed, because the used version of the dwf-deploy does not support the current environment of the dwf module. ' +
    'The VCAP_SERVICES.dwf[<n>].credentials.' + coDwfAbstract.cs_Credential_Key_Version + ' = ' + lnVersion + ' is not supported. ' +
    'Solution: 1). Use a different version of the dwf-deploy module 2). Update the environment of the dwf module';
  coValidateLib(lnVersion, 'number', lsVersionErrorMessage);// Check - version
  if (lnVersion < 1) { // Check
    throw new Error(lsVersionErrorMessage);
  }

  let laoProvidersV1 = ioCredential[coDwfAbstract.cs_Credential_Key_Providers_v1];
  coValidateLib(laoProvidersV1, 'object', // Check - providers_v1
    'Failed, because the used version of the dwf-deploy does not support the current environment of the dwf module. ' +
    'The VCAP_SERVICES.dwf[<n>].credentials.' + coDwfAbstract.cs_Credential_Key_Providers_v1 + ' = ' + laoProvidersV1 + ' is not supported. ' +
    'Solution: 1). Use a different version of the dwf-deploy module 2). Update the environment of the dwf module');

  let loRepositoryV1 = ioCredential[coDwfAbstract.cs_Credential_Key_Repository_v1];
  coValidateLib(loRepositoryV1, 'object', // Check - repository_v1
    'Failed, because the used version of the dwf-deploy does not support the current environment of the dwf module. ' +
    'The VCAP_SERVICES.dwf[<n>].credentials.' + coDwfAbstract.cs_Credential_Key_Repository_v1 + ' = ' + loRepositoryV1 + ' is not supported. ' +
    'Solution: 1). Use a different version of the dwf-deploy module 2). Update the environment of the dwf module');
  let that = this;
  this.o_sExt2aosFileExtended = {}; // contains all design time files(which needs to be deployed) with file path as string and value as {'filePath': lsFile, 'data' : data, 'hash' : sHash};
  let lo4AllFiles_sPathsHash = {}; // local object(map) for new files and their calculated hash;
  let lo4RepoFiles_sPathsHash = {}; // local object(map) for already deployed files and their hash;
  this.o_sExt2oType = {}; // Holds all the fileType we are handling
  this.o_sExt2asFile = {}; // Group the files in the module per extention
  let la_sFileUnsupported = []; // Check for unsupported extensions
  this.lbMustSync = false;
  this.loProviders = { 'POST': [], 'DELETE': [] };
  this.loObjectType = { 'POST': [], 'DELETE': [] };
  this.lmRepoCalculated = {
    'objectTypes': {},
    'objects': {},
    'ADDITION': {},
    'providers': {}
  };

  let lmProviders = iaoFilesRepo[coAbstractLib.TYPES.PROVIDERS]; // repo/providers
  let lmObjectTypes = iaoFilesRepo[coAbstractLib.TYPES.OBJECT_TYPES]; // repo/objectTypes

  /**
   *
   * @param {*} isType input string can be any of these 3 ('objectTypes','objects','ADDITION');
   * @param {*} isExt  input string name of extention
   * @param {*} isMethod  POST/DELETE but POST is considered as add and delete as minus
   * This local function keeps record or already deployed files, new post/delete and provider
   * Just before run task information generated by this method is used and then we decide
   * whether to update repo/Provider and repo/objectTypes or not
   */
  let lfUpdateMap = function (isType, isExt, isMethod) {
    let liValue = (isMethod === coAbstractLib.Methods.POST) ? 1 : -1;
    let imObj = {};
    if (that.lmRepoCalculated.hasOwnProperty(isType)) {
      imObj = that.lmRepoCalculated[isType];
      if (imObj.hasOwnProperty(isExt)) {
        imObj[isExt] = (imObj[isExt] + liValue);
      } else {
        imObj[isExt] = liValue;
      }
      that.lmRepoCalculated[isType] = imObj;
    } else {
      console.log(isType + ' didnt match any value');
    }
  };
  lmObjectTypes.forEach(function (obj) {
    lfUpdateMap(coAbstractLib.TYPES.OBJECT_TYPES, '.' + obj.ID, coAbstractLib.Methods.POST);
  });
  let lfSyncProvidersWithObjectTypes = function () {
    let lmSyncMap = {};
    lmObjectTypes.forEach(function (obj) {
      lmSyncMap[obj.PROVIDER_ID] = obj;
    });

    lmProviders.forEach(function (obj) {
      lfUpdateMap(coAbstractLib.TYPES.PROVIDERS, obj.id, coAbstractLib.Methods.POST);
      let loObjType = coAbstractLib.fGetProviderORObjectType(that.loCredential, obj.ID);
      let liobjs = (that.lmRepoCalculated[coAbstractLib.TYPES.OBJECTS].hasOwnProperty('.' + loObjType.ID)) ? that.lmRepoCalculated[coAbstractLib.TYPES.OBJECTS]['.' + loObjType.ID] : 0;
      let liAddition = (that.lmRepoCalculated.ADDITION.hasOwnProperty('.' + loObjType.ID)) ? that.lmRepoCalculated.ADDITION['.' + loObjType.ID] : 0;
      if (lmSyncMap.hasOwnProperty(obj.ID)) {
        if ((liobjs + liAddition) > 0) {
          delete lmSyncMap[obj.ID];
        } else {
          that.loProviders[coAbstractLib.PROVIDER_ACTION.DELETE].push(obj);
          that.loObjectType[coAbstractLib.PROVIDER_ACTION.DELETE].push(loObjType);
          that.lbMustSync = true;
        }

      } else {
        if ((liobjs + liAddition) > 0) {
          that.loObjectType[coAbstractLib.PROVIDER_ACTION.POST].push(loObjType);
          that.lbMustSync = true;
        } else {
          that.loProviders[coAbstractLib.PROVIDER_ACTION.DELETE].push(obj);
          that.lbMustSync = true;
        }
      }
    });
    Object.keys(lmSyncMap).forEach(function (key) {
      let loObject = lmSyncMap[key];
      let liobjs = (that.lmRepoCalculated[coAbstractLib.TYPES.OBJECTS].hasOwnProperty('.' + loObject.ID)) ? that.lmRepoCalculated[coAbstractLib.TYPES.OBJECTS]['.' + loObject.ID] : 0;
      let liAddition = (that.lmRepoCalculated.ADDITION.hasOwnProperty('.' + loObject.ID)) ? that.lmRepoCalculated.ADDITION['.' + loObject.ID] : 0;
      if ((liobjs + liAddition) > 0) {
        that.loProviders[coAbstractLib.PROVIDER_ACTION.POST].push(coAbstractLib.fGetProviderORObjectType(that.loCredential, '.' + loObject.ID));
        that.lbMustSync = true;
      } else {
        that.loObjectType[coAbstractLib.PROVIDER_ACTION.DELETE].push(loObject);
        that.lbMustSync = true;
      }
    });
  };
  //lfSyncProvidersWithObjectTypes();
  // this.loResult :: it contains all objects which needs to be deleted/post from url/repo endpoint.
  // this.loResult ::  Moreover, Deletion file list for taskchain is also got from same variable.
  this.loResult = new Result();

  /**
   * @param {*} isExt  extention of file
   * @param {*} ioData contains data of file and hash
   * function itself add the ioData to o_sExt2aosFileExtended against the isExt.
   * at the end you get all deployable files against each extention
   */
  function lfUpdateDeployableFiles(isExt, ioData) {
    let lasFile_Old = that.o_sExt2aosFileExtended[isExt];
    if (typeof lasFile_Old === 'undefined') {
      lasFile_Old = [];
    }
    let lasFile_New = lasFile_Old.concat([ioData]);
    that.o_sExt2aosFileExtended[isExt] = lasFile_New;
  }

  /*
  * @param {*} io_sExt2oValue : it contains object with key:value pairs
  * while key is 'design time directory' and value is namespace value
  * It contains files in the design time dir
  * creates a key value pair for each file with key as its 'design time file path' and value as its hash
  * name Needs to be changes
  */
  function lf_calculateHash4AllFiles(io_sExt2oValue) {
    let lsExt_Namespace = coNamespaceLib[coAbstractLib.csExtention];
    let loPreparedValue4Ext = io_sExt2oValue[lsExt_Namespace] || {};
    let laoDeployableFiles = [];
    // all extentions which needs to be checked for deployment
    let lasExts = Object.keys(that.o_sExt2oType); //[csDwfTaskChain, '.dwfnamespace', '.txt', '.test'];
    let lsFunction = coAbstractLib.cfIsDeployable;
    // loop on all extentions and check one by one whether it's deployable or not
    lasExts.forEach(function (isExtention) {
      let loType = (that.o_sExt2oType[isExtention]) ? that.o_sExt2oType[isExtention] : {};
      let lfRef = loType[lsFunction];
      if (typeof lfRef !== 'undefined') {
        let lbIsDeployable = lfRef.apply();
        //if extention is deplyoable then add all files of that extention in laoDeployableFiles.
        if (lbIsDeployable) {
          coLoggerLib.log(' * ' + isExtention + ' - deployable  ');
          //update line by removing concat :: no need for that ; just assignment is enough
          laoDeployableFiles = ((that.o_sExt2asFile[isExtention]) ? that.o_sExt2asFile[isExtention] : []);
        } else {
          coLoggerLib.log(' * ' + isExtention + ' - not deployable');
        }
      } else {
        throw new Error('Failed, because the object type for ' + isExtention + ' extention does not implement the required ' + lsFunction + ' function');
      }
      laoDeployableFiles.forEach(function (lsFile) {
        //let lsFile = file[0];
        if (coFs.statSync(lsFile).isDirectory()) { // Remove directories
        } else {
          let data = cfGetFile4BatchPost(lsFile, loPreparedValue4Ext);
          let tempObj = {};
          let sHash = cfGetHash(data);
          tempObj.data = data;

          tempObj[coAbstractLib.csKey_AOI_sHashUtf8Sha256Hex] = sHash;
          tempObj[coAbstractLib.csKey_AOI_TYPE_ID] = isExtention;
          tempObj[coAbstractLib.csKey_AOI_csNamespace] = data[coAbstractLib.csKey_AOI_csNamespace];
          tempObj[coAbstractLib.coColumn[isExtention]] = data[coAbstractLib.coColumn[isExtention]];
          tempObj[coAbstractLib.csKey_AOI_sFile] = lsFile;
          // !fasle add to if &&
          if (lo4RepoFiles_sPathsHash[lsFile] && lo4RepoFiles_sPathsHash[lsFile] === sHash && !that.loOption.overwriteFiles) {
            //if force deploy then add it to lfUpdateDeployableFiles(isExtention, tempObj);
            //lf_UpdateFile(isExtention, lsFile);
            lo4AllFiles_sPathsHash[lsFile] = sHash;
          } else if (lo4RepoFiles_sPathsHash.hasOwnProperty(lsFile) && lo4RepoFiles_sPathsHash[lsFile] !== sHash) {
            tempObj[coAbstractLib.csKey_AOI_Action] = coAbstractLib.Methods.POST;
            lfUpdateDeployableFiles(isExtention, tempObj);
            lo4AllFiles_sPathsHash[lsFile] = lo4RepoFiles_sPathsHash[lsFile];
            lfUpdateMap('ADDITION', isExtention, coAbstractLib.Methods.POST);
          } else {
            tempObj[coAbstractLib.csKey_AOI_Action] = coAbstractLib.Methods.POST;
            lfUpdateDeployableFiles(isExtention, tempObj);
            lo4AllFiles_sPathsHash[lsFile] = lo4RepoFiles_sPathsHash[lsFile];
            lfUpdateMap('ADDITION', isExtention, coAbstractLib.Methods.POST);
          }
        }
      });
      laoDeployableFiles = [];
    });
  }

  /*
  * It is used for already deployed files
  * creates a key value pair for each file with key as its 'design time file path' and value as its hash
  */
  function lf_saveHash4DeployedFiles() {
    that.laoFilesRepo.forEach(function (currentFile) {
      lo4RepoFiles_sPathsHash[currentFile.DESIGN_TIME_FILE_PATH] = currentFile.HASH_UTF8_SHA256_HEX;
      lfUpdateMap(coAbstractLib.TYPES.OBJECTS, currentFile.TYPE_ID, coAbstractLib.Methods.POST);
    });
  }
  lf_saveHash4DeployedFiles();

  // Supported extention to Type and checked uniqueness
  let lfPush_o_sExt2oType = function (ioType) {
    let lsExt = ioType[coAbstractLib.csExtention];
    let loType = that.o_sExt2oType[lsExt];
    if (typeof loType === 'undefined') {
      that.o_sExt2oType[lsExt] = ioType;
    } else {
      throw new Error('Failed, because the ' + lsExt + ' extention is already supported by ' +
        JSON.stringify(loType) + ' and cannot be supported also by ' + JSON.stringify(ioType));
    }
  };
  lfPush_o_sExt2oType(coTxtLib);
  lfPush_o_sExt2oType(coNamespaceLib);
  lfPush_o_sExt2oType(coTaskchainLib);
  lfPush_o_sExt2oType(coDLMProfileLib);

  let lfPush_o_sExt2asFile = function (isExt, isFile) {
    let lasFile_Old = that.o_sExt2asFile[isExt];
    if (typeof lasFile_Old === 'undefined') {
      lasFile_Old = [];
    }
    let lasFile_New = lasFile_Old.concat([isFile]);
    that.o_sExt2asFile[isExt] = lasFile_New;
  };

  /*
  * This function takes isEXt as extention and isFile as filePah
  * Deletes isFile with extention isExt from  laoFilesRepo  if file is found
  * At the end any file remained in laoFilesRepo mean those  files has to be deleted from runtime.
  */
  function lf_UpdateFile(isExt, isFile) {
    let lasFile_Old = that.o_sExt2asFile[isExt];
    if (typeof lasFile_Old === 'undefined') {
      return;
    }
    lasFile_Old.splice(lasFile_Old.indexOf(isFile), 1);   // removing first from o_sExt2asFile which is base container for files
    that.o_sExt2asFile[isExt] = lasFile_Old;
    /*
    if(NoDelete) {
      //add the below loop to this if statement;
    }
    */
    //removing from array that we created from Repo/objects get call
    that.laoFilesRepo.forEach(function (obj) {
      if (obj.DESIGN_TIME_FILE_PATH === isFile) {
        that.laoFilesRepo.splice(that.laoFilesRepo.indexOf(obj), 1);
      }
    });
  }
  this.laoFileDesign.forEach(function (isFile) {
    let lsFile = isFile[0];
    if (coFs.statSync(lsFile).isDirectory()) { // Remove directories
      return;
    }
    let sExtnameLowerCase = coPathLib.extnameNamelessFile(lsFile).toLowerCase(); // Group files
    lfPush_o_sExt2asFile(sExtnameLowerCase, lsFile);

  });
  // Check for unsupported extensions
  Object.getOwnPropertyNames(this.o_sExt2asFile).forEach(function (isExt) {
    let laoFilePerExt = that.o_sExt2asFile[isExt];
    if (typeof that.o_sExt2oType[isExt] === 'undefined') { // Unsupported
      la_sFileUnsupported = la_sFileUnsupported.concat(laoFilePerExt);
    }
  });
  if (la_sFileUnsupported.length > 0) {
    throw new Error('Failed, because the ' + JSON.stringify(la_sFileUnsupported) + ' files in the module are not supported');
  }
  this.loResultPrepare = this.prepare();
  if (this.loResultPrepare[coAbstractLib.csKey_nStatus] === coAbstractLib.cnSuccess) {

    let loResultPrepare_o_sExt2oValue = this.loResultPrepare[coAbstractLib.csKey_o_sExt2oValue];
    lf_calculateHash4AllFiles(loResultPrepare_o_sExt2oValue);
    /*
    * loop on deployedFiles map to check if it matches with new files(design time)
    * on every match it calls lf_UpdateFile to delete from the correcsponding map object
    */
    let updateFiles = function () {
      Object.keys(lo4RepoFiles_sPathsHash).forEach(function (key) {
        if (lo4AllFiles_sPathsHash.hasOwnProperty(key)) {
          if (lo4RepoFiles_sPathsHash[key] === lo4AllFiles_sPathsHash[key]) {
            let sExtnameLowerCase = coPathLib.extnameNamelessFile(key).toLowerCase();
            lf_UpdateFile(sExtnameLowerCase, key);
          }
        }
      });
    };
    updateFiles();// called in the lf_calculateAllFiels
    this.cfPreparePostBatchFiles();
    this.laoFilesRepo.forEach(function (currentFile) {
      lfUpdateMap('ADDITION', currentFile.TYPE_ID, coAbstractLib.Methods.DELETE);
    });
    lfSyncProvidersWithObjectTypes();
  }
}

/**
 * @param {*} isExt extention of file (i: incoming, s: string, Ext, Extention, all together isExt)
 *
 */
Container.prototype.lfGetDeployableFiles = function (isExt) {
  let lasFile_Old = this.o_sExt2aosFileExtended[isExt];
  if (typeof lasFile_Old === 'undefined') {
    lasFile_Old = [];
  }
  return lasFile_Old;
};

Container.prototype.cfPreparePostBatchFiles = function () {
  let that = this;
  let loObjectProperties = {};
  let loItem = {};
  let lsExt = '';
  let laoAllFilesToDeploy = [];
  Object.keys(that.o_sExt2aosFileExtended).forEach(function (key) {
    let laoToDeploy = that.o_sExt2aosFileExtended[key];
    lsExt = key;
    let extentionHandler = that.o_sExt2oType[lsExt];
    if (extentionHandler) {
      let lRef = extentionHandler[coAbstractLib.cfCommonObjectProperties];
      if (lRef) {
        loObjectProperties = lRef.apply();
      } else {
        return;
      }
    } else {
      return;
    }
    //loObjectProperties = (coAbstractLib.coExtentions.hasOwnProperty(lsExt)) ? coAbstractLib.coExtentions[lsExt] : {};
    laoAllFilesToDeploy = laoToDeploy;
    laoToDeploy.forEach(function (obj) {
      loItem[coAbstractLib.csKey_AOI_Action] = obj[coAbstractLib.csKey_AOI_Action]; // introducing new object property with POST/DELETE ACTION VALUE
      loItem[coAbstractLib.csKey_AOI_nStatus] = coAbstractLib.cnSuccess;
      loItem[coAbstractLib.csKey_AOI_sDescription] = '';
      Object.keys(loObjectProperties).forEach(function (currentProperty) {
        loItem[loObjectProperties[currentProperty]] = obj[loObjectProperties[currentProperty]];
      });
      //update this log becasue it can crash when obj[coAbstractLib.csKey_AOI_sFile] is undefined
      coLoggerLib.log(' * ' + lsExt + ' - ' + obj[coAbstractLib.csKey_AOI_sFile] + ' ... ok (to ' + obj[coAbstractLib.csKey_AOI_Action] + ')');
      that.loResult.addItem(lsExt, loItem);
      loItem = {};
    });
  });
  /**
   * loops over all files which are filtered after comparison with new files
   * creates an object with all info of file and set the flaf of Action as Delete
   * finally add each object in the array
   */
  // !dontDelete
  if (!that.loOption.dontDeleteFiles) {
    that.laoFilesRepo.forEach(function (isFile) {
      let lsFile = isFile;
      let lsExt = lsFile[coAbstractLib.csKey_AOI_TYPE_ID];
      let extentionHandler = that.o_sExt2oType[lsExt];
      if (extentionHandler) {
        let lRef = extentionHandler[coAbstractLib.cfCommonObjectProperties];
        if (lRef) {
          loObjectProperties = lRef.apply();
        } else {
          return;
        }
      } else {
        return;
      }
      loItem[coAbstractLib.csKey_AOI_Action] = coAbstractLib.Methods.DELETE; // introducing new object property with POST/DELETE ACTION VALUE
      loItem[coAbstractLib.csKey_AOI_nStatus] = coAbstractLib.cnSuccess;
      loItem[coAbstractLib.csKey_AOI_sDescription] = '';
      if (!loObjectProperties) {
        return;
      }
      Object.keys(loObjectProperties).forEach(function (key) {
        loItem[loObjectProperties[key]] = lsFile[key];
      });
      coLoggerLib.log(' * ' + lsExt + ' - ' + lsFile.DESIGN_TIME_FILE_PATH + ' ... ok (to ' + loItem[coAbstractLib.csKey_AOI_Action] + ')');
      that.loResult.addItem(lsExt, loItem);
      //updateMap('ADDITION', lsExt, coAbstractLib.Methods.DELETE);
      loItem = {};
    });
  }
};

Container.prototype.prepare = function () {
  let that = this;
  let lsFunction = coAbstractLib.cfPrepare;
  let loResult = new Result();
  Object.keys(that.o_sExt2oType).forEach(function (isExt) {
    let loType = that.o_sExt2oType[isExt];
    let lasFile = that.o_sExt2asFile[isExt] || []; // Check - empty src folder, only .txt or .dwfnamespace ot .dwftaskchain
    let lfRef = loType[lsFunction];
    if (typeof lfRef !== 'undefined') {
      let loResultExt = lfRef.apply(that, [lasFile, that.loOption]);
      loResult.addResult(loResultExt);
    }
  });
  return loResult;
};

Container.prototype.createTasks = function () {
  let that = this;
  let lsFunction = coAbstractLib.cfCreateTasks;
  let loResult = new Result();
  //let lbPostProvider = coAbstractLib.PROVIDER_ACTION.SKIP;
  Object.keys(that.o_sExt2oType).forEach(function (isExt) {
    let lsExt = isExt;

    // Here we use data generated by lfUpdateMap to decide the repo/provider and repo/objectType calls
    let liAlreadyDeployed = (that.lmRepoCalculated[coAbstractLib.TYPES.OBJECTS].hasOwnProperty(lsExt)) ? that.lmRepoCalculated[coAbstractLib.TYPES.OBJECTS][lsExt] : 0;
    let liNewDeployed = (that.lmRepoCalculated.ADDITION.hasOwnProperty(lsExt)) ? that.lmRepoCalculated.ADDITION[lsExt] : 0;
    if (liAlreadyDeployed >= 0 && (liAlreadyDeployed + liNewDeployed) > 0) {
      if (!that.lmRepoCalculated[coAbstractLib.TYPES.OBJECT_TYPES].hasOwnProperty(lsExt)) {
        that.loObjectType.POST.push(coAbstractLib.fGetObjectType(that.loCredential, lsExt));
        let loProvider = coAbstractLib.fGetProvider(that.loCredential, lsExt);
        if (!that.lmRepoCalculated[coAbstractLib.TYPES.PROVIDERS].hasOwnProperty(loProvider.ID)) {
          that.loProviders.POST.push(loProvider);
        }
        that.lbMustSync = true;
      }
    }
    let loType = that.o_sExt2oType[lsExt];
    let lfRef = loType[lsFunction];
    if (typeof lfRef !== 'undefined') {
      let laoToUpdate = (that.loResult.o_sExt2aoItem[lsExt]) ? that.loResult.o_sExt2aoItem[lsExt] : [];
      if ((laoToUpdate && laoToUpdate.length > 0) || that.lbMustSync) {
        //that.loProviders = coAbstractLib.fGetProviders(that.loCredential, lbPostProvider, lsExt, that.loProviders);
        //that.loObjectType = coAbstractLib.fGetObjectTypes(that.loCredential, lbPostProvider, lsExt, that.loObjectType);

        let loResultExt = lfRef.apply(that, [that.lfGetDeployableFiles(lsExt), laoToUpdate, that.loCredential, that.loOption, that.loProviders, that.loObjectType]);
        loResult.addResult(loResultExt);

        that.lbMustSync = false;
        that.loProviders = { 'POST': [], 'DELETE': [] };
        that.loObjectType = { 'POST': [], 'DELETE': [] };

      } else {
        coLoggerLib.log(` * ${lsExt} - no designtime files to deploy nor update`);
      }
    }
  });
  return loResult;
};

Container.prototype.runTasks = function (io_sExt2afTask, ifCb) {
  let lafTasks = [];
  let that = this;
  Object.keys(that.o_sExt2oType).forEach(function (isExt) {
    let lafTasks4Ext = io_sExt2afTask[isExt];
    if (typeof lafTasks4Ext !== 'undefined') {
      lafTasks = lafTasks.concat(lafTasks4Ext);
    }
  });

  coAsyncLib.series(lafTasks, ifCb);
};

module.exports.Container = Container;
