'use strict';

const https = require('https');
const xsenv = require('@sap/xsenv');
var logger = require('./lib/logger.js');
logger.timerInit('overall');

var pjson = require('./package.json');
var options = require('./lib/options.js');
var args = require('./lib/arguments.js');

const coProcessEnvLib = require('./lib/dwf/processEnv.js');
const coProcessEnv = new coProcessEnvLib.ProcessEnv();

var opt = options();

var exit_rc0_silent = Error('');

// try to import ssl certificates
try {
  let laoCertificate = xsenv.loadCertificates();
  if (laoCertificate) {
    https.globalAgent.options.ca = laoCertificate;
  } else {
    // Note - perform that before processing the command arguments, to ensure their processing in the right order
    let lnValue = coProcessEnv.setNodeTlsRejectUnauthorizedToAccept(); // TODO - Shall disallow calls to untrusted systems
    logger.warn('Sets the value for the process.env.NODE_TLS_REJECT_UNAUTHORIZED to ' + lnValue + ', because the xsenv did return \'' + laoCertificate + '\' certificates.');
  }
} catch (error) {
  throw error;
}

function idle() {
  setTimeout(idle, 10 * 60 * 1000);
}

function signalProcessExit(exitCode, showOverallTime, forcedExit) {
  if (showOverallTime) {
    logger.log(logger.timerDelta('overall'));
  }

  if (opt.exit || exitCode || forcedExit) {
    var written = process.stdout.write("\n");
    if (written) {
      // stdout was already drained, just exit now
      process.exit(exitCode);
    } else {
      // register for drain event, and idle until then
      process.stdout.once('drain', function () {
        process.exit(exitCode);
      });
      // fall through to idle()
    }
  } else {
    logger.log('Application can be stopped.');
    // fall through to idle()
  }

  // always idle() if we get here
  idle();

  // note: returning to caller here, idle() is handled in the event loop
}

// set defaults from environment variables
if (typeof process.env.EXIT !== "undefined") {
  opt.exit = true;
}

// check environment variable for more options and translate them to process arguments
const DWF_DEPLOY_OPTIONS = 'DWF_DEPLOY_OPTIONS';
if (typeof process.env[DWF_DEPLOY_OPTIONS] !== "undefined") {
  var options;
  try {
    options = JSON.parse(process.env[DWF_DEPLOY_OPTIONS]);
  } catch (error) {
    logger.error('Failed to parse JSON object in ' + DWF_DEPLOY_OPTIONS + ': ' + error);
    process.exit(1);
  }

  for (var option in options) {
    if (option === 'verbose') {
      args.translateJSONEnvBooleanOptionToOption(logger, options, DWF_DEPLOY_OPTIONS, option, process.argv);
    } else {
      logger.error('Unknown option ' + option + ' in ' + DWF_DEPLOY_OPTIONS);
      process.exit(1);
    }
  }
}

// process arguments
for (var i = 2; i < process.argv.length; ++i) {
  var arg = process.argv[i];
  if (arg === '-h' || arg === '--help') {
    var help = [
      "",
      "  " + pjson.name + ", version " + pjson.version,
      "",
      "  Options:",
      "",
      "    -h, --help                    print usage information and exit",
      "        --version                 print version and exit",
      "    -t, --trace                   enable tracing",
      "        --[no-]verbose            [don't] print detailed log messages to the console",
      "",
      "        --[no-]exit               [don't] exit after deployment of artifacts",
      "                                  by default, the application will not exit, but enter an idle mode",
      "",
      "        --[no-]reject-unauthorized",
      "                                  [don't] reject unauthorized http endpoints by default",
      "    --[no-]auto-undeploy          Does not delete runtime files missing in Design time Directory",
      "    --[force-]deploy              Deploys files even if they exist already in runtime (overwrites)",
      ""
    ];
    logger.log(help.join("\n"));
    signalProcessExit(0, false, true);
    return;
  } else if (arg === '--version') {
    logger.log(pjson.version);
    signalProcessExit(0, false, true);
    return;
  } else if (arg === '-t' || arg === '--trace') {
    logger.setTrace(true);
  } else if (arg === '--verbose') {
    opt.verbose = true;
  } else if (arg === '--no-verbose') {
    opt.verbose = false;
  } else if (arg === '--exit') {
    opt.exit = true;
  } else if (arg === '--no-exit') {
    opt.exit = false;
  } else if (arg === '--reject-unauthorized') {
    coProcessEnv.setNodeTlsRejectUnauthorizedToReject();
  } else if (arg === '--no-reject-unauthorized') {
    coProcessEnv.setNodeTlsRejectUnauthorizedToAccept();
  } else if (arg === '--force-deploy') {
    opt.overwriteFiles = true;
  } else if (arg === '--deploy') {
    opt.overwriteFiles = false;
  }else if (arg === '--no-auto-undeploy') {
    opt.dontDeleteFiles = true;
  } else if (arg === '--auto-undeploy') {
    opt.dontDeleteFiles = false;
  }  else {
    logger.error('Unknown argument: ' + arg);
    process.exit(1);
  }
}

logger.log(pjson.name + '... v.' + pjson.version);
// trace the options object after processing the arguments
logger.trace('options:', opt);
// trace the relevant process.env
logger.trace('processEnv:', coProcessEnv.getObject());

// argument processing is complete

// apply the logger specific options
logger.setVerbose(opt.verbose);
logger.setLogFile(opt.logFile);

var async = require('async');
var services;

function initializeServices(cb) {
  try {
    // read services and service replacements
    services = require('./lib/services.js')();

    cb();
  } catch (error) {
    cb(error.message);
  }
}

function processFilesAndDeploy(cb) {
  try {
    // finally, do the dirty work
    var deploy = require('./lib/deploy.js');
    deploy(opt, services, cb);
  } catch (error) {
    cb(error.message);
  }
}

var tasks = [
  initializeServices,
  processFilesAndDeploy
];

async.series(tasks, function (err) {
  if (err === exit_rc0_silent) {
    signalProcessExit(0, false, true);
  } else {
    if (err) {
      if (err.message) {
        logger.error('Error: ' + err.message);
      } else {
        logger.error('Error: ' + err);
      }
      if (err.stack) {
        logger.trace('Stack: ' + err.stack);
      }
    }
    signalProcessExit(err ? 1 : 0, true);
  }
});
